#! /usr/bin/env python

import os, sys
from myexceptions import *
from linenum import LN

# Everything in this module would be much easier if MAXSTACK in parser.h of
# the Python source code was much higher than 500. Nested tuples can only
# be read to depth 34.

NONE, LP, RP, NODENAME, CTOKEN, LEAF = range(6)
CodeNames = ('NONE', 'LP', 'RP', 'NODENAME', 'CTOKEN', 'LEAF')
OPEN = '<<OPEN>>'
CLOSE = '<<CLOSE>>'
NEWLINE = '<<NEWLINE>>'
TAB = '<<TAB>>'
SpecTokens = (NEWLINE, TAB, OPEN, CLOSE)

def CheckTokenListTypes(alist):
    """Lists the token codes in a tuple of tokens."""
    types = dict()

    def recurse(seq):
        for item in seq:
            t = type(item)
            if t in (tuple, list):
                recurse(item)
            else:
                if t not in types:
                    types[t] = None

    recurse(alist)
    if len(types) == 0:
            return 'empty'
    elif len(types) == 1:
        if Token in types:
            return 'Token'
        if str in types:
            return 'str'
    else:
        return types.keys()

def TokenTuple_AsString(tup):
    """Represents a tuple of tokens as an informative string."""

    def recurse(seq):
        out = []
        for item in seq:
            if isinstance(item, tuple):
                substring = '(%s)' % recurse(item)
                out.append(substring)
            elif isinstance(item, list):
                substring = '[%s]' % recurse(item)
                out.append(substring)
            else:
                if isinstance(item, Token):
                    item = item.asstring()
                out.append(repr(item))
        if isinstance(seq, tuple):
            return '(%s)' % ', '.join(out)
        elif isinstance(seq, list):
            return '[%s]' % ', '.join(out)
        else:
            return ', '.join(out)

    s = CheckTokenListTypes(tup)
    text = recurse(tup)
    return '[of "%s": %s]' % (s, text)
                
def Tuple_ToTokenTuple(text):
    """Converts an input string in ".data" file format into a tuple of tokens."""
    text = text.strip()
    words = text.splitlines()
    lastword = None
    out = []
    for word in words:
        if lastword == OPEN:
            tok = Token(word, NODENAME)
        elif word == OPEN:
            tok = Token(word, LP)
        elif word == CLOSE:
            tok = Token(word, RP)
        else:
            tok = Token(word, CTOKEN)
        out.append(tok)
        lastword = word
    return tuple(out)

class Token(object):
    """Contains token data."""
    counter = 0

    def __init__(self, token_string, token_code):
        """Create a Token by calling Token(token_string, token_code)"""
        self.id = Token.counter
        Token.counter += 1
        self.string = token_string
        self.code = token_code
        self.parent = None
        self.children = list()
        self.left = None
        self.right = None
        self.depth = 0
        self.walkaround_code = NONE

    def copy(self):
        """Copies a token."""
        return Token(self.string, self.code)

    def __str__(self):
        return '<Token %i: |%s|, %s>' % \
              (self.id, self.string, CodeNames[self.code])

    def __repr__(self):
        if self.code == 'OP':
            template = "'%s'"
        else:
            template = '"\'%s\'"'
        s = template % self.string
        if self.parent is None:
            p = 'None'
        else:
            p = self.parent.id
        if self.left is None:
            ls = 'None'
        else:
            ls = self.left.id
        if self.right is None:
            rs = 'None'
        else:
            rs = self.right.id
        return '<Token %i (%s, %s): parent %s, lsib %s, rsib %s, pass %s>' % \
             (self.id, s, self.code, p, ls, rs, self.walkaround_code)

    def which_child_am_I(self):
        """Finds index of a Token in the list of its parent's children."""
        if self.parent is None:
            raise ValueError, 'top Token has no parent'
        for i, child in enumerate(self.parent.children):
            if child == self:  # Default "__cmp__" used: memory addresses.
                return i

    def reset(self):
        """Reinitializes a Token."""
        self.parent = None
        self.left = None
        self.right = None
        self.children = []
        self.depth = 0
        self.walkaround_code = NONE

    def asstring(self):
        """Simplest sting describing a Token."""
        if self.code == 'OP':
            template = '%s'
        else:
            template = "'%s'"
        return template % self.string

    def is_open_parens(self):
        """Does this token contain a '('?"""
        return self.string == OPEN and self.code == LP

    def is_close_parens(self):
        """Does this token contain a ')'?"""
        return self.string == CLOSE and self.code == RP

    def equiv(token):
        """Do two Tokens contain the same string and code."""
        return self.string == token.string and self.num == token.num

    def isleaf(self):
        """Is this Token a leaf?"""
        return len(self.children) == 0
