#! /usr/local/bin/python

import os, sys
from myexceptions import *
from linenum import LN

GnuCEmi_base = 'GnuCEmitter'
GnuCEmi_print_base = 'GnuCEmitter_print'
GnuCEmi_grammar = GnuCEmi_base + '.g'
GnuCEmi_print_grammar = GnuCEmi_print_base + '.g'
GnuCEmi_print_java = GnuCEmi_print_base + '.java'
GnuCEmi_grammar_fixed = GnuCEmi_base + '.g.fixed'
copied_file = os.path.join('../grammars', GnuCEmi_print_grammar)
java_compile_dir = os.path.abspath('../grammars')

errfile = 'insert_prints_errors.txt'

"""Modifies "GnuCEmitter.g.fixed" so both the parse tree and the tokens are printed. 

               ./insert_prints.py [<preproc>]

    where the optional <preproc> is "FULL", "TOKENONLY", or "NONE". The
    output is copied to the 'grammars" directory and a "make" is done there.

1. Modifications to "GnuCEmitter.g"

GnuCEmitter.g needs to be regularized slightly:
    Remove spaces at the end of lines.
    Rules should be separated by a single blank line.
    Remove blank lines from inside rules.
    Remove the four spaces in front of "unaryOperator".
    Replace each action of the form
        print( variable );
    by
        <@ variable @>
    Replace tabs with 8 spaces.
    Remove commented out material.
    Move the colon in "attributeDecl:" to the next line.
It would be nicer to do this with the ANTLR grammar grammar.

I have made many changes in the java code part of "GnuCEmitter.g.fixed". Run
a "diff" to find them.

2. Introduction

There are two places where this program inserts print statements. Here is
a rule from GnuCEmitter.g:

1 typeQualifier
2         :       a:"const"                       { print( a ); }
3         |       b:"volatile"                    { print( b ); }
4         ;

Here is the same rule in "GnuCEmitter.g.fixed":

1 typeQualifier
2         :       a:"const"                       { <@ a @> }
3         |       b:"volatile"                    { <@ b @> }
4         ;

"insert_prints.py" modifies this rule in four places.

1  typeQualifier
2    { if ( inputState.guessing==0 ) {
3            print(Open);
4            print("typeQualifier");
5         }
6      }
7          :  (
8                  a:"const"                       {  print(Open); print("typeQualifier.0"); print( a ); print(Close); }
9          |       b:"volatile"                    {  print(Open); print("typeQualifier.1"); print( b ); print(Close); }
10            )
11   { currentOutput.print(Close + MyTokenSep); }
12         ;

Lines 2-5 and 11 cause
    <<OPEN>>
    typeQualifier
    ...
    <<CLOSE>>
to be printed whenever the rule is triggered. The parentheses in lines 7 and
10 are added to ensure that line 10 is executed. The parentheses create a
subrule.

Line 8 causes
    <<OPEN>>
    typeQualifier.0
    const
    <<CLOSE>>
to be printed whenever the token "const" is found. Line 9 is similar.
Lines 8 and 9 are called "token print"s.
"""
#>==================================================================

NONE, TOKENONLY, FULL = 0, 1, 2

def constant(text, do_preprocs):
    return text.replace('<@PREPROC@>', str(do_preprocs))

# This class brings together the code that processes the text for a rule.
class RulePrints:
    """Processes rule prints"""
    # When inputState.guessing != 0, "antlr" is walking through a tree
    # looking for a match. "antlr" doesn't execute any actions in this case
    # except actions at the beginning of a rule. Only the successful match
    # should be printed. See The ANTLR Reference Manual, ANTLR Meta-Language,
    # Syntactic Predicates.

    # Template for rule print at start of rule.
    templateR1 = \
"""  { if ( inputState.guessing==0 ) {
          print(Open);
          print("%s");
       }
    }"""

    # Template for the special case of initDecl.
    templateR2 = \
"""  { if ( inputState.guessing==0 ) {
          print(Open);
          print("%s");
       }
    %s"""

    # Template for rule print at end of rule.
    templateR3a = '   )'
    templateR3b = """\
  { currentOutput.print(Close + MyTokenSep); }
        ;"""

    # Special end-of-rule template for externalList (no comma at end).
    templateR4a = '   )'
    templateR4b = """\
  { currentOutput.print(Close); }
        ;"""

    def __init__(self):
        pass

    def AddParens(self, line):
        """Add parentheses around a rule.

        We want to convert everything after the ":" in the rule to a single
        subrule. Then we can put a { ... } after the subrule and it will be
        executed.
        """
        j = 0
        while line[j] == ' ':
            j += 1
        line1 = line[:j+1] + '  ('
        line2 = '         ' + line[j+1:]
        return [line1, line2]

    # Note: this function makes in-place changes to "lines".
    def ProcessRule(self, lines):
        """Wraps a rule print around a rule"""
        if lines[0] == 'protected':
            ruleline = 1
        else:
            ruleline = 0
        rule_name = lines[ruleline].strip()

        # A few rules start with "protected" instead of the rule name.
        if lines[0] == 'protected':
            lines[2:2] = [RulePrints.templateR1 % rule_name]
            lines[3:4] = self.AddParens(lines[3])
            lines[-1] = lines[-1][:-1] + RulePrints.templateR3a
            lines.extend(RulePrints.templateR3b.splitlines())
            lines[1] = rule_name
        elif rule_name == 'initDecl':
            # {...}{...} is not allowed. This problem only occurs for "initDeclk".
            line1 = lines[1].strip()[2:]
            lines[1:2] = [RulePrints.templateR2 % (rule_name, line1)]
            lines[2:3] = self.AddParens(lines[2])
            lines[-1] = lines[-1][:-1] + RulePrints.templateR3a
            lines.extend(RulePrints.templateR3b.splitlines())
            lines[0] = rule_name
        elif rule_name == 'externalList':
            lines[1:1] = [RulePrints.templateR1 % rule_name]
            lines[2:3] = self.AddParens(lines[2])
            lines[-1] = lines[-1][:-1] + RulePrints.templateR4a
            lines.extend(RulePrints.templateR4b.splitlines())
            lines[0] = rule_name
        else:
            lines[1:1] = [RulePrints.templateR1 % rule_name]
            lines[2:3] = self.AddParens(lines[2])
            lines[-1] = lines[-1][:-1] + RulePrints.templateR3a
            lines.extend(RulePrints.templateR3b.splitlines())
            lines[0] = rule_name
        return rule_name, ruleline

class TokenPrints:
    """Processes token prints."""
    templateT = ' print(Open); print("%s"); print(%s); print(Close);'
    templateS = """ print(Open); print("stringConst.0"); print( s );
                print(Close);"""
    templateAD = """// Processes strings without using stringConst.
                  print(Open); print("attributeDecl.1"); print( b );
                  print(Close);"""

    def __init__(self):
        pass

    def ProcessLine(self, line, rule_name, count):
        """Each <@ ... @> pair in a line is found and processed."""
        end = 0
        start = line.find('<@', end)
        if start == -1:
            return count, line
        parts = []
        while 1:
            parts.append(line[end:start])
            end = line.find('@>', start)
            if end == -1:
                raise ValueError, 'missing "@>" in line:\n%s' % line
            end += 2  # Pass the "@>".
            var = line[start+2:end-2]
            newname = '%s.%i' % (rule_name, count)
            count += 1
            if rule_name == 'attributeDecl' and var.strip() == 'b':
               newprint = TokenPrints.templateAD
            elif rule_name == 'stringConst':
                newprint = TokenPrints.templateS
            else:
                newprint = TokenPrints.templateT % (newname, var)
            parts.append(newprint)
            start = line.find('<@', end)
            if start == -1:
                break
        parts.append(line[end:])
        aline = ''.join(parts)
        return count, aline

def DoRule(text, ruleprints, tokenprints):
    """Processes rule prints and token prints for some rule."""
    # Much of this should be done using the ANTLR grammar grammar.
    rules = text.split('\n\n')
    out = []
    for rule in rules:
        lines = rule.splitlines()
        try:
            rule_name, ruleline = ruleprints.ProcessRule(lines)
        except:
            print >> sys.stderr, 'Error processing rule:'
            print >> sys.stderr, '----'
            print >> sys.stderr, '\n'.join(lines)
            print >> sys.stderr, '----'
            raise
    
        count = 0
        out2 = []
        for line in lines:
            count, lineout = tokenprints.ProcessLine(line, rule_name, count)
            out2.append(lineout)
        text = '\n'.join(out2)
        out.append(text)
    out.append('')
    return '\n\n'.join(out)    

def SplitFile(text):
    """Split "GnuCEmitter.g.fixed" into two parts.

    Grammar files start with a bunch of java followed by the grammar itself. 
    Separate the two.
    """
    splitpoint = text.find('\nexternalList') + 1
    top = text[:splitpoint]
    # In Java, filename == class name.
    top = top.replace(GnuCEmi_base, GnuCEmi_print_base)
    bottom = text[splitpoint:].strip()
    return top, bottom

#>=== Main program =================================================

def doit(text, do_preprocs):
    """Does all the work."""
    top, bottom = SplitFile(text)
    top = constant(top, do_preprocs)
    ruleprints = RulePrints()
    tokenprints = TokenPrints()
    bottomout = DoRule(bottom, ruleprints, tokenprints)
    sys.stdout.flush()
    text = '\n\n'.join( (top, bottomout) )
    return text

def makeit(text):
    """Makes "GnuCEmitter_print.g" in the "grammars" directory."""
    if os.path.isfile(errfile):
        os.remove(errfile)

    # Put a copy of the output in directory "grammars".
    file(copied_file, 'w').write(text)

    # Do a make in directory "grammars".
    command = 'cd %s; make &> %s' % \
          (java_compile_dir, os.path.abspath(errfile))
    os.system(command)

def filestuff():
    """sys.argv and file stuff."""

    if len(sys.argv) not in (1, 2):
        sys.exit('USAGE: ./insert_prints.py [<preprocs>] where ' \
            'preprocs may be "FULL", "TOKENONLY", or "NONE".')


    if len(sys.argv) == 1:
       do_preprocs = NONE
    else:
        if sys.argv[1] == 'FULL':
           do_preprocs = FULL
        elif sys.argv[1] == 'TOKENONLY':
           do_preprocs = TOKENONLY
        elif sys.argv[1] == 'NONE':
           do_preprocs = NONE
        else:
            sys.exit('argument must be "FULL", "TOKENONLY", or "NONE".')

    text = file(GnuCEmi_grammar_fixed, 'r').read()
    text = doit(text, do_preprocs)
    file(GnuCEmi_print_grammar, 'w').write(text)

    makeit(text)

    if os.path.isfile(errfile):
        text = file(errfile, 'r').read()
        for s in ('\n' + GnuCEmi_print_grammar + ':',
                  '\n' + GnuCEmi_print_java + ':'):
            if s in text:
                raise JavaException, 'java error. See "%s".' % errfile
        os.remove(errfile)

if __name__ == '__main__':
    filestuff()
