#! /usr/bin/env python

import os, sys
import ctokens
from ctokens import NONE, LP, RP, NODENAME, CTOKEN, LEAF, CodeNames, \
   NEWLINE, TAB, OPEN, CLOSE, SpecTokens
import strip_preprocs, unstrip_preprocs
from myexceptions import *
from linenum import LN

# Everything in this module would be much easier if MAXSTACK in parser.h of
# the Python source code was much higher than 500. Nested tuples can only
# be read to depth 34.

#>=== Globals ====================================================

# This is the list of nodenames to keep when using Pyrex to wrap C extensions.
keeplist = ('externalList', 'declaration', 'storageClassSpecifier',
            'structSpecifier', 'unionSpecifier', 'structOrUnionBody',
            'structOrUnionBody.0', 'structOrUnionBody.2',
            'structOrUnionBody.4', 'structDeclaration', 'enumSpecifier',
            'enumSpecifier.1', 'enumList', 'enumerator', 'declarator.0',
            'declarator.1', 'declarator.2')

#>=== class Tree =================================================

class Tree(object):
    """Tree representation for a parser's output."""
    def __init__(self, data=None):
        """Tree is created using Tree(data), data optional."""
        self.top = None
        if isinstance(data, str):
            data = ctokens.Tuple_ToTokenTuple(data)
        if isinstance(data, (tuple, list)):
            self.fromtuple(data)
        elif isinstance(data, ctokens.Token):
            self.addtoken(data, 'TOP')
        elif data is not None:
            raise TypeError, '"data" must be a string, list, tuple, Token, or None'

    def __str__(self):
        if self.top is None:
            return ''
        self.depths()
        out = []
        for token in self.walk():
            if not isinstance(token, ctokens.Token):
                raise TypeError, 'node must be a Token'
            text = '%s%s' % (token.depth * '  ', str(token))
            out.append(text)
        return '\n'.join(out)

    def __repr__(self):
        if self.top is None:
            return ''
        self.depths()
        out = []
        for token in self.walk():
            text = '%s%s' % (token.depth * '  ', repr(token))
            out.append(text)
        return '\n'.join(out)

    def show_walkaround(self):
        """Like "__str__" but with "walkaround" instead of "walk"."""
        self.depths()
        out = []
        for token in self.walkaround():
            text = '%s%s' % (token.depth * '  ', str(token))
            out.append(text)
        return '\n'.join(out)

    def show_reverse_walk(self):
        """Like "__str__" but with "reverse_walk" instead of "walk"."""
        self.depths()
        out = []
        for token in self.reverse_walk():
            text = '%s%s' % (token.depth * '  ', str(token))
            out.append(text)
        return '\n'.join(out)

    def show_parens(self):
        """Like "__str__" but the '('s and ')'s are added."""
        self.depths()
        out = []
        for token in self.walkaround():
            inner = []
            inner.append(token.depth * '  ')
            if token.walkaround_code == LP:
                inner.append('(')
            inner.append(str(token))
            if token.walkaround_code == RP:
                inner.append(')')
            text = ''.join(inner)
            out.append(text)
        return '\n'.join(out)

    def copy(self):
        """Copies a tree and all of its Tokens."""
        if self.top is None:
            return tree()
        newtree = Tree(self.top.copy())

        def recurse(oldtoken, newtoken):
            newtoken.children = []
            for child in oldtoken.children:
                newchild = child.copy()
                newtree.addtoken(newchild, newtoken, 'LAST')
                if not child.isleaf():
                    recurse(child, newchild)

        recurse(self.top, newtree.top)
        return newtree

    def walk(self, starttoken=None):
        """Generator that does top-down, left-right traversal of tree."""
        if self.top is None:
            return
        if starttoken is None:
            starttoken = self.top
        yield starttoken
        for child in starttoken.children:
            for token in self.walk(child):
                yield token
                
    def reverse_walk(self, starttoken=None):
        """Generator that does top-down, left-right traversal of tree.

        Use this if you plan to delete nodes.
        """
        if self.top is None:
            return
        if starttoken is None:
            starttoken = self.top
        r = len(starttoken.children)
        for i in range(r-1, -1, -1):
            for token in self.reverse_walk(starttoken.children[i]):
                yield token
        yield starttoken

    def walkaround(self, starttoken=None):
        """Walks around the outside of a tree.

        Each NODENAME node is visited twice. The first visit is before its
        children are visited and corresponds to '('. At the start of the
        first visit, self.walkaround_code is set to LP. The second visit is
        after its children are visited and corresponds to ')'. Just before
        the second visit, self.walkaround_code is set to RP. Some NODENAMEs
        may be leaves. They are still visited twice.
        """
        if self.top is None:
            return
        if starttoken is None:
            starttoken = self.top
        if starttoken.code == NODENAME:
            starttoken.walkaround_code = LP
            yield starttoken
            for child in starttoken.children:
                for token in self.walkaround(child):
                    yield token
            starttoken.walkaround_code = RP
            yield starttoken
            starttoken.walkaround_code = NONE
        else:
            starttoken.walkaround_code = LEAF
            yield starttoken
            starttoken.walkaround_code = NONE

    def depths(self):
        """Assigns each node the depth it has in the tree."""
        if self.top is None:
            return
        self.top.depth = 0
        for token in self.walk():
            if token.parent is not None:
                token.depth = token.parent.depth + 1

    def indexes(self):
        """Assigns each node its index among its parent's children."""

        def recurse(token):
            for i, token2 in enumerate(token.children):
                token2.index = i
                if not token2.isleaf():
                    recurse(token2)

        if self.top is None:
            return
        self.top.index = None
        recurse(self.top)

    def fromtuple(self, tokentuple):
        """Adds a tuple of Tokens to a new Tree."""
        if self.top is not None:
            raise ValueError, 'new or reset Tree required'
        if len(tokentuple) == 0:
            return
        i = 0
        tt = tokentuple  # Shorter name.
        # "child_bearer" is the node that children are currently being added
        # to.
        child_bearer = 'TOP'
        while i < len(tt):
            if tt[i].is_open_parens():
                i += 1  # Skip the '('.
                self.addtoken(tt[i], child_bearer)
                child_bearer = tt[i]
                i += 1
            elif tt[i].is_close_parens():
                if child_bearer == self.top:
                     return
                else:
                    child_bearer = child_bearer.parent
                i += 1
            else:  # C tokens.
                self.addtoken(tt[i], child_bearer)
                i +=1

    def addtoken(self, token, parent, index='LAST'):
        """Add a Token to a Tree.

        "parent" can be a Token, or the string "TOP".
        "index" can be either an int or the string "LAST".
        If a token is added, it keeps its children.
        """
        if not isinstance(token, ctokens.Token):
            raise TypeError, '"token" must be a Token'
        if parent == 'TOP':
            self.top = token
            token.parent = None
            token.left = None
            token.right = None
        elif isinstance(parent, ctokens.Token):
            if index != 'LAST' and \
                   (index < 0 or index > len(parent.children)):
                raise IndexError, '"index" out of range'
            token.parent = parent

            # Set the left sibling.
            if index == 'LAST':
                index = len(parent.children)
            if index == 0 or len(parent.children) == 0:
                token.left = None
            else:
                token.left = parent.children[index-1]
                token.left.right = token

            # Set the right sibling.
            if index == len(parent.children) or len(parent.children) == 0:
                token.right = None
            else:
                token.right = parent.children[index]
                token.right.left = token
            parent.children.insert(index, token)
        else:
            raise TypeError, '"parent" must be a Token or the string "TOP"'

    def add_child_list(self, newchildren, parent, index='LAST'):
        """Add a properly linked list of children to a node.

        "parent" can be a Token, or the string "TOP". "index" can be either
        an int or the string "LAST". Children that are added keep their
        children. "newchildren" must be properly linked to each other as a
        set of siblings.
        """
        if not isinstance(parent, ctokens.Token):
            raise TypeError, '"parent" must be a Token'
        if not isinstance(newchildren, (list)):
            raise TypeError, '"newchildren" must be a list'
        if index != 'LAST' and not isinstance(index, int):
            raise TypeError, '"index" must be an int or the string "LAST". It has %s' % \
                 type(index)
        if index < 0 or index > len(parent.children):
            raise IndexError, '"index" out of range'
        if index == 'LAST':
            index = len(parent.children)
        if len(newchildren) == 0:
            return
        for child in newchildren:
            child.parent = parent

        # Set the left sibling.
        if index == 0 or len(parent.children) == 0:
            newchildren[0].left = None
        else:
            newchildren[0].left = parent.children[index-1]
            newchildren[0].left.right = newchildren[0]

        # Set the right sibling.
        if index == len(parent.children) or len(parent.children) == 0:
            newchildren[-1].right = None
        else:
            newchildren[-1].right = parent.children[index]
            newchildren[-1].right.left = newchildren[-1]

        parent.children[index:index] = newchildren

    def remove_subtree(self, token):
        """Remove a Token from the tree. Its children go with it."""
        if self.top == token:
            self.top = None
            return Tree(token)
        elif isinstance(token, ctokens.Token):
            if token.left is not None:
                token.left.right = token.right
                token.left = None
            if token.right is not None:
                token.right.left = token.left
                token.right = None
            token.parent.children.remove(token)
            return Tree(token)
        else:
            raise TypeError, '"token" must be a Token'

    def killtoken(self, token):
        """Removes a token from the tree. It is replaced by its children.

        A copy of "token" is returned.
        """
        if token is self.top:
            self.top = None
        elif token.isleaf():
            self.remove_subtree(token)
        else:
            index = token.which_child_am_I()
            parent = token.parent
            children = token.children
            tree = self.remove_subtree(token)
            self.add_child_list(children, parent, index)
        return token.copy()

    def combine(self):
        """Combines adjacent leaf nodes."""
        for token in self.reverse_walk():
            if token.isleaf():
                continue
            for i in range(len(token.children)-1, 0, -1):
                if token.children[i].code == CTOKEN and \
                       token.children[i-1].code == CTOKEN and \
                       token.children[i].string not in SpecTokens and \
                       token.children[i-1].string not in SpecTokens:
                    tok = token.children[i]
                    self.remove_subtree(token.children[i])
                    token.children[i-1].string = \
                         '%s %s' % (token.children[i-1].string, tok.string)

    def strip_special_tokens(self):
        """Removes '<<NEWLINE>>' and '<<TAB>>' Tokens from a Tree."""
        for token in self.reverse_walk():
            if not token.isleaf():
                continue
            if token.string in SpecTokens:
                self.killtoken(token)

    def stripNODENAMEleaves(self):
        """Remove leaves that are also NODENAMEs.

        Use this function only after the tree is completed.
        """
        for token in self.reverse_walk():
            if token.isleaf() and token.code == NODENAME:
                self.killtoken(token)

    def select(self, selectdata, keep):
        """Removes selected tokens from a tree.

         "selectdata" can be a list, tuple, dict, or "RULEPRINTSONLY".
        """
        def will_keep(keep, nodename, selectdata):
            if selectdata == 'RULEPRINTSONLY':
                 base, ext = os.path.splitext(nodename)
                 return len(ext) == 0 or not ext[1:].isdigit()
            else:
                return (keep and nodename in selectdata) or \
                       (not keep and nodename not in selectdata)

        for token in self.reverse_walk():
            if not token.isleaf() and \
                    not will_keep(keep, token.string, selectdata):
                if token == self.top:
                    token.string = None
                else:
                    self.killtoken(token)

    def to_string_list(self):
        """Produces a list of the stings in the Tokens of a Tree."""
        out = []
        first = True
        for token in self.walk():
            if not first:
                out.append(token.string)
            else:
                first = False
        return out

    def make_source(self):
        """Make compilable C code from a Tree.

        Modifies the Tree.
        """
        first, special, normal = 0, 1, 2
        self.stripNODENAMEleaves()
        self.select([], True)
        strings = self.to_string_list()
        out = []
        prev = first
        for i, string in enumerate(strings):
            if string == NEWLINE:
                out.append('\n')
                prev = special
            elif string == TAB:
                out.append('\t')
                prev = special
            else:
                if prev == normal:
                    out.append(' ')
                prev = normal
                out.append(string)
        out.append('\n')
        return ''.join(out)

    def reconstruct_text(self, LPisNone_sep, LPtoken_sep, LPindent,
              token_sep, indent, RPtoken_sep, RPindent, indent_size,
              evalable=False, suppress_special_tokens=False):
        """Converts a Tree into a useful and/or pretty string.

        LPisNone_sep   If the flattened form would start with "((",
                       put "LPisNone_sep" between the "("s.
        LPtoken_sep    Goes before all other "(",s.
        LPindent       If True, put indentation after "LPtoken_sep".
        token_sep      Separates token strings.
        indent         If True, put indentation after "token_sep".
        RPtoken_sep    Goes before all  ")",s.
        RPindent       If True, put indentation after "RPtoken_sep".
        indent_size    Number of spaces per indent.
        evalable       Output will be eval'able.
        suppress_special_tokens  If True, skip all the "<<NEWLINE>>" and
                                 "<<TAB>>" tokens

        An example is very useful. Create "tree" and run:

            tree.reconstruct_text(LPisNone_sep = '\nAA',
                  LPtoken_sep = ',\nBB', LPindent = True,
                  token_sep = ', \nCC', indent = True,
                  RPtoken_sep = 'DD', RPindent = False,
                  indent_size = 2, evalable = False,
                  suppress_special_tokens = False)

        The output is:

            (externalList,
            BB  (externalDef,
            BB    (declaration,
            BB      (declSpecifiers,
            BB        (typeSpecifier,
            BB          (typeSpecifier.3,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |int|DD)DD)DD),
            BB      (initDeclList,
            BB        (initDecl,
            BB          (declarator,
            BB            (declarator.0,
            CC              |k|DD)DD)DD)DD),
            CC      |;|DD)DD),
            ...
        """
        if self.top is None:
            return ''
        self.depths()
        self.indexes()
        topsNone = self.top.string is None
        counter = 0
        out = []
        for token in self.walkaround():
            if token.walkaround_code == LP:
                # We want nothing to be before the initial "(", so the
                # counter == 0 case is ignored.
                if counter == 1 and topsNone:
                    out.append(LPisNone_sep)
                elif counter > 0:
                    out.append(LPtoken_sep)
                if LPindent: 
                    spaces = token.depth * indent_size * ' '
                    if not topsNone or counter != 1 or \
                               LPisNone_sep[-1] == '\n':
                        out.append(spaces)
                out.append('(')
                if token.string is not None:
                    if evalable:
                        out.append(repr(token.string))
                    else:
                        out.append(token.string)
            elif token.walkaround_code == RP:
                out.append(RPtoken_sep)
                if RPindent:
                    out.append(token.depth * indent_size * ' ')
                out.append(')')
            elif not suppress_special_tokens or \
                      token.string not in (NEWLINE, TAB):
                out.append(token_sep)
                if indent:
                    out.append(token.depth * indent_size * ' ')
                if evalable:
                    out.append(repr(token.string))
                else:
                    out.append("|%s|" % token.string)
            counter += 1
        out.append('\n')
        return ''.join(out)

    def asstring_pretty(self, flatRP=True, flatCtoken=True,
              evalable=False, suppress_special_tokens=False):
        """Tree pretty printer."""
        _LPisNone_sep = '\n'
        _LPtoken_sep = ',\n'
        _LPindent = True

        if flatRP:
            _RPtoken_sep = ''
            _RPindent = False
        else:
            _RPtoken_sep = '\n'
            _RPindent = True

        if flatCtoken:
            _token_sep = ', '
            _indent = False
        else:
            _token_sep = ',\n'
            _indent = True
        _indent_size = 2

        return self.reconstruct_text(_LPisNone_sep,
                  _LPtoken_sep, _LPindent,
                  _token_sep, _indent,
                  _RPtoken_sep,  _RPindent,
                  _indent_size, evalable,
                  suppress_special_tokens)

    def asstring_flat(self, suppress_special_tokens=False):
        """Outputs a string which is the repr of Python nested tuple.

        Because the Python parser stack is too small, eval will usually
        fail on this string.
        """
        return self.reconstruct_text(LPisNone_sep = '',
                  LPtoken_sep = ', ', LPindent = False,
                  token_sep = ', ', indent = False,
                  RPtoken_sep = '', RPindent = False,
                  indent_size = 0, evalable=True,
                  suppress_special_tokens = suppress_special_tokens)

    def asstring_test(self, suppress_special_tokens=False):
        """Demo for "reconstruct_text"."""
        return self.reconstruct_text(LPisNone_sep = '\nAA',
                  LPtoken_sep = ',\nBB', LPindent = True,
                  token_sep = ', \nCC', indent = True,
                  RPtoken_sep = 'DD', RPindent = False,
                  indent_size = 2, evalable=False,
                  suppress_special_tokens = suppress_special_tokens)

    def to_datafile_format(self):
        """Outputs a string which is in the formal for a ".data" file."""
        out = []
        for token in self.walkaround():
            if token.walkaround_code == LP:
                out.append(OPEN)
                out.append(token.string)
            elif token.walkaround_code == RP:
                out.append(CLOSE)
            elif token.walkaround_code == LEAF:
                out.append(token.string)
        out.append('')
        return '\n'.join(out)

#>=== Functions that work with Trees ==================================

def Combine(tree):
    """Combines adjacent leaf nodes. Returns a new tree."""
    return tree.copy().combine()

def StripSpecialTokens(tree):
    """Removes '<<NEWLINE>>' and '<<TAB>>' Tokens from a Tree. Returns a new tree."""
    return tree.copy().trip_special_tokens()

def StripNODENAME_Leaves(tree):
    """Remove leaves that are also NODENAMEs. Returns a new tree.

    Use this function only after the tree is completed.
    """
    return tree.copy().stripNODENAMEleaves()

def Select(tree, selectdata, keep):
    """Removes selected tokens from a tree. A new tree is returned.

    "selectdata" can be a list, tuple or dict.
    """
    return tree.copy().select(selectdata, keep)

def MakeSource(tree):
    """Make compilable C code from a Tree. Returns a new tree."""
    return tree.copy().make_source()

def MakeNest(tree):
    return tree.asstring_pretty()

def MakeNest_Files(datafile, nestfile):
    """Outputs the data from a ".data" file in a easily readable form."""
    text = file(datafile, 'r').read()
    tree = Tree(text)
    text = tree.asstring_pretty()
    file(nestfile, 'w').write(text)

def MakeSrc_Files(datafile, srcfile):
    """Outputs the data from a ".data" file as compilable C code."""
    text = file(datafile, 'r').read()
    tree = Tree(text)
    text = tree.make_source()
    file(srcfile, 'w').write(text)

def SelectPyrex(tree):
    """Selects using "keeplist". Returns a new tree."""
    selectdict = dict()
    for word in keeplist:
        selectdict[word] = None
    tree = tree.copy()
    tree.stripNODENAMEleaves()
    tree.select(selectdict, True)

def Run(filename):
    """Does end-to-end processing of a C file.

    The five stages are: strip_preprocs, UnparsePlus, makenest, makesrc, and
    unstrip_preprocs.
    """
    stripfile = '%s.strip' % filename
    ppsfile = '%s.pps' % filename
    datafile = '%s.data' % filename
    errfile = '%s.err' % filename
    nestfile = '%s.nest' % filename
    tmpsrcfile = '%s.tmpsrc' % filename
    srcfile = '%s.src' % filename
    temperr = 'temp.err'
    if not os.path.isfile(filename):
        raise IOError, 'file not found: %s' % filename

    # Strip out the preprocessor directives.
    strip_preprocs.stripit(filename, stripfile, ppsfile)
    if os.path.exists(errfile):
        os.remove(errfile)

    # Run "UnparsePlus.java". This creates a ".data" file.
    if os.path.exists(temperr):
        os.remove(temperr)
    command = 'java UnparsePlus %s %s %s &> %s' % \
          (stripfile, datafile, errfile, temperr)
    os.system(command)
    if os.path.exists(temperr):
        f = file(temperr, 'r')
        text = f.read()
        f.close()
        os.remove(temperr)
        if len(text) > 0:
            raise JavaError, text
    if os.path.exists(errfile):
        if os.stat(errfile).st_size == 0:
            os.remove(errfile)
        else:
            f = file(errfile, 'r')
            text = f.read()
            f.close()
            os.remove(errfile)
            raise JavaException, text

    # Read in the ".data" file.
    text = file(datafile, 'r').read()

    # Convert it into a tree.
    tree = Tree(text)

    # Create the ".nest" file.
    textout = MakeNest(tree)
    file(nestfile, 'w').write(textout)

    # Reconstruct the source code and put it in a ".tmpsrc" file.
    textout = tree.make_source()  # I don't care if I mess up the tree.
    file(tmpsrcfile, 'w').write(textout)

    # Put the preprocessor directives back in.
    unstrip_preprocs.unstripit(ppsfile, tmpsrcfile, srcfile)

if __name__ == '__main__':
    if len(sys.argv) not in (2, 3):
        sys.exit('USAGE: ./tree.py <filename> [<dowhat>] ' \
           'where <dowhat> can be "nest" or "source"')
    if len(sys.argv) == 3:
        text = file(filename + '.data', 'r').read()
        tree = Tree(text)
        if sys.argv[2] == 'nest':
            textout = MakeNest(tree)
            file(nestfile, 'w').write(textout)
        elif sys.argv[2] == 'source':
            textout = tree.make_source()
            file(tmpsrcfile, 'w').write(textout)
        else:
            sys.exit('If present, the second argument must be ' \
                       '"nest" or "source"')
    else:
        Run(sys.argv[1])
